[pushed] c++: elaborated-type-spec in requires-expr [PR101677]

Message ID 20220405162639.2098812-1-jason@redhat.com
State Committed
Commit 1de6612d994ada8edaab18bbc6afd8e9a57413aa
Headers
Series [pushed] c++: elaborated-type-spec in requires-expr [PR101677] |

Commit Message

Jason Merrill April 5, 2022, 4:26 p.m. UTC
  We were failing to declare class S in the global namespace because we were
treating the requires-expression parameter scope as a normal block scope, so
the implicit declaration went there.

It seems to me that the requires parameter scope is more like a function
parameter scope (not least in the use of the word "parameter"), so let's
change the scope kind.  But then we need to adjust the prohibition on
placeholders declaring implicit template parameters.

Tested x86_64-pc-linux-gnu, applying to trunk.

	PR c++/101677

gcc/cp/ChangeLog:

	* name-lookup.h (struct cp_binding_level): Add requires_expression
	bit-field.
	* name-lookup.cc (pushtag): Check it.
	* parser.cc (cp_parser_requires_expression): Set it.

gcc/testsuite/ChangeLog:

	* g++.dg/cpp2a/concepts-requires28.C: New test.
---
 gcc/cp/name-lookup.h                             |  5 ++++-
 gcc/cp/parser.cc                                 |  5 +++--
 gcc/testsuite/g++.dg/cpp2a/concepts-pr67178.C    |  2 +-
 gcc/testsuite/g++.dg/cpp2a/concepts-requires28.C | 13 +++++++++++++
 4 files changed, 21 insertions(+), 4 deletions(-)
 create mode 100644 gcc/testsuite/g++.dg/cpp2a/concepts-requires28.C


base-commit: 5c8d22b00adedc21f8b697dd6b990f4821a06634
  

Patch

diff --git a/gcc/cp/parser.cc b/gcc/cp/parser.cc
index 7e1c777364e..bfd16e1ef62 100644
--- a/gcc/cp/parser.cc
+++ b/gcc/cp/parser.cc
@@ -29964,7 +29964,8 @@  cp_parser_requires_expression (cp_parser *parser)
       scope_sentinel ()
       {
 	++cp_unevaluated_operand;
-	begin_scope (sk_block, NULL_TREE);
+	begin_scope (sk_function_parms, NULL_TREE);
+	current_binding_level->requires_expression = true;
       }
 
       ~scope_sentinel ()
@@ -48082,7 +48083,7 @@  static tree
 synthesize_implicit_template_parm  (cp_parser *parser, tree constr)
 {
   /* A requires-clause is not a function and cannot have placeholders.  */
-  if (current_binding_level->kind == sk_block)
+  if (current_binding_level->requires_expression)
     {
       error ("placeholder type not allowed in this context");
       return error_mark_node;
diff --git a/gcc/testsuite/g++.dg/cpp2a/concepts-pr67178.C b/gcc/testsuite/g++.dg/cpp2a/concepts-pr67178.C
index c74f6f00a5c..bdd5f9699e0 100644
--- a/gcc/testsuite/g++.dg/cpp2a/concepts-pr67178.C
+++ b/gcc/testsuite/g++.dg/cpp2a/concepts-pr67178.C
@@ -12,7 +12,7 @@  concept C0 = requires (auto x) { // { dg-error "placeholder type" }
 template<typename T>
 concept C1 = requires (C1 auto x) { // { dg-error "not been declared|placeholder|two or more|in requirements" }
   x; // { dg-error "not declared" }
-  { x } -> c; // { dg-message "is invalid" }
+  { x } -> c; // { dg-message "is invalid|not declared" }
 };
 
 template<typename T>
diff --git a/gcc/testsuite/g++.dg/cpp2a/concepts-requires28.C b/gcc/testsuite/g++.dg/cpp2a/concepts-requires28.C
new file mode 100644
index 00000000000..e632f0167f4
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp2a/concepts-requires28.C
@@ -0,0 +1,13 @@ 
+// PR c++/101677
+// { dg-do compile { target c++20 } }
+
+template<class T>
+concept C_bug_with_forward_decl = requires(T& t){
+    t.template f<class S>();
+};
+
+struct good {
+    template<class T> void f() {}
+};
+
+static_assert(C_bug_with_forward_decl<good>);
diff --git a/gcc/cp/name-lookup.h b/gcc/cp/name-lookup.h
index 68d08725a00..fa039028847 100644
--- a/gcc/cp/name-lookup.h
+++ b/gcc/cp/name-lookup.h
@@ -309,7 +309,10 @@  struct GTY(()) cp_binding_level {
   /* true for SK_FUNCTION_PARMS of immediate functions.  */
   unsigned immediate_fn_ctx_p : 1;
 
-  /* 22 bits left to fill a 32-bit word.  */
+  /* True for SK_FUNCTION_PARMS of a requires-expression.  */
+  unsigned requires_expression: 1;
+
+  /* 21 bits left to fill a 32-bit word.  */
 };
 
 /* The binding level currently in effect.  */